use super::{db::Noose, migrations::MigrationRunner};
use crate::{
    bussy::{channels, Command, Message, PubSub},
    utils::{config::Config as ServiceConfig, error::Error, structs::Subscription},
};
use async_trait::async_trait;
use deadpool_sqlite::{Config, Object, Pool, Runtime};
use nostr::{nips::nip01::Coordinate, Event, EventId, Filter, RelayMessage, Timestamp, Url};
use nostr_database::{Backend, DatabaseOptions, NostrDatabase, Order};
use rusqlite::Row;
use sea_query::{extension::sqlite::SqliteExpr, Order as SqOrder, Query, SqliteQueryBuilder};
use sea_query_rusqlite::RusqliteBinder;
use std::sync::Arc;
use std::{collections::HashSet, str::FromStr};

#[derive(Debug, Clone)]
struct EventRow {
    id: String,
    pubkey: String,
    created_at: i64,
    kind: i64,
    tags: String,
    sig: String,
    content: String,
}

impl From<&Row<'_>> for EventRow {
    fn from(row: &Row) -> Self {
        let id: String = row.get("id").unwrap();
        let pubkey: String = row.get("pubkey").unwrap();
        let created_at: i64 = row.get("created_at").unwrap();
        let kind: i64 = row.get("kind").unwrap();
        let tags: String = row.get("tags").unwrap();
        let sig: String = row.get("sig").unwrap();
        let content: String = row.get("content").unwrap();

        Self {
            id,
            pubkey,
            created_at,
            kind,
            tags,
            sig,
            content,
        }
    }
}

impl From<&EventRow> for Event {
    fn from(row: &EventRow) -> Self {
        row.to_event()
    }
}

impl EventRow {
    pub fn to_event(&self) -> Event {
        let tags: Vec<Vec<String>> = serde_json::from_str(&self.tags).unwrap();

        let event_json = serde_json::json!(
            {
                "id": self.id,
                "content": self.content,
                "created_at": self.created_at,
                "kind": self.kind,
                "pubkey": self.pubkey,
                "sig": self.sig,
                "tags": tags
            }
        );

        Event::from_value(event_json).unwrap()
    }
}

enum EventsTable {
    Table,
    EventId,
    Kind,
    Pubkey,
    Content,
    CreatedAt,
    Tags,
    Sig,
}

impl sea_query::Iden for EventsTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "events",
                Self::EventId => "id",
                Self::Kind => "kind",
                Self::Pubkey => "pubkey",
                Self::Content => "content",
                Self::CreatedAt => "created_at",
                Self::Tags => "tags",
                Self::Sig => "sig",
            }
        )
        .unwrap()
    }
}

enum EventsFTSTable {
    Table,
    EventId,
    Content,
}

impl sea_query::Iden for EventsFTSTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "events_fts",
                Self::EventId => "id",
                Self::Content => "content",
            }
        )
        .unwrap()
    }
}

enum TagsTable {
    Table,
    Tag,
    Value,
    EventId,
}

impl sea_query::Iden for TagsTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "tags",
                Self::Tag => "tag",
                Self::Value => "value",
                Self::EventId => "event_id",
            }
        )
        .unwrap()
    }
}

enum DeletedCoordinatesTable {
    Table,
    Coordinate,
    CreatedAt,
}

impl sea_query::Iden for DeletedCoordinatesTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "deleted_coordinates",
                Self::Coordinate => "coordinate",
                Self::CreatedAt => "created_at",
            }
        )
        .unwrap()
    }
}

enum EventSeenByRelaysTable {
    Table,
    Id,
    EventId,
    RelayURL,
}

impl sea_query::Iden for EventSeenByRelaysTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "event_seen_by_relays",
                Self::Id => "id",
                Self::EventId => "event_id",
                Self::RelayURL => "relay_url",
            }
        )
        .unwrap()
    }
}

#[derive(Debug)]
pub struct NostrSqlite {
    pool: Pool,
}

impl NostrSqlite {
    pub async fn new(config: Arc<ServiceConfig>) -> Self {
        let env_db_path = std::env::var("DATABASE_URL").unwrap_or("/tmp/sqlite.db".to_string());

        let cfg = Config::new(env_db_path);
        let pool = cfg.create_pool(Runtime::Tokio1).unwrap();

        if NostrSqlite::run_migrations(&pool).await {
            return Self { pool };
        }

        panic!("Pool's closed");
    }

    async fn run_migrations(pool: &Pool) -> bool {
        let connection = pool.get().await.unwrap();
        connection.interact(MigrationRunner::up).await.unwrap()
    }

    async fn get_connection(&self) -> Result<Object, Error> {
        match self.pool.get().await {
            Ok(connection) => Ok(connection),
            Err(err) => Err(Error::internal(err.into())),
        }
    }

    async fn save_event(&self, event: &Event) -> Result<bool, Error> {
        let event = event.clone();

        let id = event.id.to_string();
        let kind = event.kind.as_u64();
        let pubkey = event.pubkey.to_string();
        let content = event.content.to_string();
        let created_at = event.created_at.as_i64();
        let tags = serde_json::to_string(&event.tags).unwrap();
        let sig = event.sig.to_string();

        let ids: Vec<String> = event.event_ids().map(|eid| eid.to_string()).collect();

        if self.event_is_too_old(&event) {
            return Ok(false);
        }

        if event.is_ephemeral() {
            return Ok(false);
        }

        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let Ok(event_saved) = connection
            .interact(move |conn| -> Result<bool, Error> {
                let tx = conn.transaction().unwrap();

                if event.is_replaceable() {
                    dbg!("new event is replaceable - searching for previously stored event");
                    let (sql, values) = Query::select()
                        .from(EventsTable::Table)
                        .columns([EventsTable::EventId])
                        .and_where(sea_query::Expr::col(EventsTable::Pubkey).eq(&pubkey))
                        .and_where(sea_query::Expr::col(EventsTable::Kind).eq(kind))
                        .and_where(sea_query::Expr::col(EventsTable::CreatedAt).gte(created_at))
                        .limit(1)
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Ok(res) = tx.execute(sql.as_str(), &*values.as_params()) {
                        if res > 0 {
                            return Ok(true);
                        }
                    };
                    
                }

                if event.is_parameterized_replaceable() {
                    dbg!(
                    "new event is parametrized replaceable - searching for previously stored event"
                );
                    let d_tags: Vec<String> = event
                        .tags
                        .iter()
                        .filter(|tag| tag.kind() == nostr::TagKind::D)
                        .map(|tag| tag.clone().to_vec()[1].clone())
                        .collect();
                    let (sql, values) = Query::select()
                        .from(EventsTable::Table)
                        .column((EventsTable::Table, EventsTable::EventId))
                        .left_join(
                            TagsTable::Table,
                            sea_query::Expr::col((TagsTable::Table, TagsTable::EventId))
                                .equals((EventsTable::Table, EventsTable::EventId)),
                        )
                        .and_where(
                            sea_query::Expr::col((EventsTable::Table, EventsTable::Pubkey))
                                .eq(&pubkey),
                        )
                        .and_where(
                            sea_query::Expr::col((EventsTable::Table, EventsTable::Kind)).eq(kind),
                        )
                        .and_where(sea_query::Expr::col((TagsTable::Table, TagsTable::Tag)).eq("d"))
                        .and_where(
                            sea_query::Expr::col((TagsTable::Table, TagsTable::Value))
                                .eq(d_tags[0].to_string()),
                        )
                        .and_where(
                            sea_query::Expr::col((EventsTable::Table, EventsTable::CreatedAt))
                                .gte(created_at),
                        )
                        .limit(1)
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Ok(results) = tx.execute(sql.as_str(), &*values.as_params()) {
                        if results > 0 {
                            return Ok(true);
                        }
                    }
                }

                // Insert replaceble event
                if event.is_replaceable() {
                    dbg!("deleting older replaceable event from events table");
                    let (sql, values) = Query::delete()
                        .from_table(EventsTable::Table)
                        .and_where(
                            sea_query::Expr::col((EventsTable::Table, EventsTable::Kind)).eq(kind),
                        )
                        .and_where(
                            sea_query::Expr::col((EventsTable::Table, EventsTable::Pubkey))
                                .eq(&pubkey),
                        )
                        .and_where(
                            sea_query::Expr::col((EventsTable::Table, EventsTable::EventId))
                                .not_in_subquery(
                                    Query::select()
                                        .from(EventsTable::Table)
                                        .column(EventsTable::EventId)
                                        .and_where(sea_query::Expr::col(EventsTable::Kind).eq(kind))
                                        .and_where(
                                            sea_query::Expr::col(EventsTable::Pubkey).eq(&pubkey),
                                        )
                                        .order_by(EventsTable::CreatedAt, sea_query::Order::Desc)
                                        .limit(1)
                                        .to_owned(),
                                ),
                        )
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Ok(results) = tx.execute(sql.as_str(), &*values.as_params()) {
                        if results > 0 {
                            log::info!(
                                "removed {} older replaceable kind {} events for author: {:?}",
                                results,
                                event.kind.as_u32(),
                                event.pubkey.to_string()
                            );
                        }
                    }
                }

                // Insert parametrized replaceble event
                if event.is_parameterized_replaceable() {
                    dbg!("deleting older parametrized replaceable event from events table");
                    log::info!("deleting older parametrized replaceable event from events table");
                    let d_tag = event.identifier();
                    let (sql, values) = Query::delete()
                        .from_table(EventsTable::Table)
                        .and_where(
                            sea_query::Expr::col((EventsTable::Table, EventsTable::EventId))
                                .in_subquery(
                                    Query::select()
                                        .from(EventsTable::Table)
                                        .column((EventsTable::Table, EventsTable::EventId))
                                        .left_join(
                                            TagsTable::Table,
                                            sea_query::Expr::col((
                                                TagsTable::Table,
                                                TagsTable::EventId,
                                            ))
                                            .equals((EventsTable::Table, EventsTable::EventId)),
                                        )
                                        .and_where(
                                            sea_query::Expr::col((
                                                EventsTable::Table,
                                                EventsTable::Kind,
                                            ))
                                            .eq(kind),
                                        )
                                        .and_where(
                                            sea_query::Expr::col((
                                                EventsTable::Table,
                                                EventsTable::Pubkey,
                                            ))
                                            .eq(&pubkey),
                                        )
                                        .and_where(
                                            sea_query::Expr::col((
                                                TagsTable::Table,
                                                TagsTable::Tag,
                                            ))
                                            .eq("d"),
                                        )
                                        .and_where(
                                            sea_query::Expr::col((
                                                TagsTable::Table,
                                                TagsTable::Value,
                                            ))
                                            .eq(d_tag),
                                        )
                                        .to_owned(),
                                ),
                        )
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Ok(results) = tx.execute(sql.as_str(), &*values.as_params()) {
                        if results > 0 {
                            log::info!("removed {} older parameterized replaceable kind {} events for author: {:?}", results, event.kind, event.pubkey);
                        }
                    }
                }

                if event.kind == nostr::Kind::EventDeletion {
                    // Delete from Events
                    let (sql, values) = Query::delete()
                        .from_table(EventsTable::Table)
                        .and_where(sea_query::Expr::col(EventsTable::Kind).ne(5))
                        .and_where(sea_query::Expr::col(EventsTable::Pubkey).eq(&pubkey))
                        .and_where(sea_query::Expr::col(EventsTable::EventId).is_in(&ids))
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Err(err) = tx.execute(sql.as_str(), &*values.as_params()) {
                        tx.rollback().unwrap();

                        return Ok(false);
                    }

                    // Delete from EventsFTS
                    let (sql, values) = Query::delete()
                        .from_table(EventsFTSTable::Table)
                        .and_where(sea_query::Expr::col(EventsFTSTable::EventId).is_in(ids))
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Err(err) = tx.execute(sql.as_str(), &*values.as_params()) {
                        tx.rollback().unwrap();

                        return Ok(false);
                    }
                } else {
                    log::debug!("inserting new event in events");
                    // Insert into Events table
                    let (sql, values) = Query::insert()
                        .into_table(EventsTable::Table)
                        .columns([
                            EventsTable::EventId,
                            EventsTable::Content,
                            EventsTable::Kind,
                            EventsTable::Pubkey,
                            EventsTable::CreatedAt,
                            EventsTable::Tags,
                            EventsTable::Sig,
                        ])
                        .values_panic([
                            id.clone().into(),
                            content.clone().into(),
                            kind.into(),
                            pubkey.into(),
                            created_at.into(),
                            tags.into(),
                            sig.into(),
                        ])
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Err(err) = tx.execute(sql.as_str(), &*values.as_params()) {
                        log::error!("Error inserting event into 'events' table: {}", err);
                        tx.rollback().unwrap();

                        return Ok(false);
                    };

                    // Insert into EventsFTS table
                    log::debug!("inserting new event into eventsFTS");
                    let (sql, values) = Query::insert()
                        .into_table(EventsFTSTable::Table)
                        .columns([EventsFTSTable::EventId, EventsFTSTable::Content])
                        .values_panic([id.clone().into(), content.into()])
                        .build_rusqlite(SqliteQueryBuilder);

                    if let Err(err) = tx.execute(sql.as_str(), &*values.as_params()) {
                        log::error!("Error inserting event into 'eventsFTS' table: {}", err);
                        tx.rollback().unwrap();

                        return Ok(false);
                    }

                    // Insert into Tags table
                    log::debug!("inserting new event into tags");
                    for tag in event.tags.clone() {
                        let tag = tag.to_vec();
                        if tag.len() >= 2 {
                            let tag_name = &tag[0];
                            let tag_value = &tag[1];
                            if tag_name.len() == 1 {
                                let (sql, values) = Query::insert()
                                    .into_table(TagsTable::Table)
                                    .columns([TagsTable::Tag, TagsTable::Value, TagsTable::EventId])
                                    .values_panic([
                                        tag_name.into(),
                                        tag_value.into(),
                                        id.clone().into(),
                                    ])
                                    .build_rusqlite(SqliteQueryBuilder);

                                if let Err(err) = tx.execute(sql.as_str(), &*values.as_params()) {
                                    log::error!("Error inserting event into 'tags' table: {}", err);
                                    tx.rollback().unwrap();

                                    return Ok(false);
                                }
                            }
                        }
                    }
                }

                match tx.commit() {
                    Ok(_) => Ok(true),
                    Err(err) => {
                        log::error!("Error during transaction commit: {}", err);
                        Ok(false)
                    }
                }
            })
            .await else {
                return Err(Error::internal_with_message("Failed to execute query"));
        };

        event_saved
    }

    async fn has_event_already_been_saved(&self, event_id: &EventId) -> Result<bool, Error> {
        match self.get_event_by_id(*event_id).await {
            Ok(_) => Ok(true),
            Err(_) => Ok(false),
        }
    }

    async fn has_event_id_been_deleted(&self, event_id: &EventId) -> Result<bool, Error> {
        match self.get_event_by_id(*event_id).await {
            Ok(_) => Ok(false),
            Err(_) => Ok(true),
        }
    }

    /// Skip events that are older than 10 minutes
    fn event_is_too_old(&self, event: &Event) -> bool {
        if chrono::Utc::now().timestamp() - 600 > event.created_at.as_i64() {
            return true;
        }
        false
    }

    async fn set_event_id_seen(&self, event_id: EventId, relay_url: Url) -> Result<(), Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let event_id = event_id.to_string();
        let relay_url = relay_url.to_string();

        let Ok(query_result) = connection
            .interact(|conn| -> Result<usize, Error> {
                let (sql, values) = Query::insert()
                    .into_table(EventSeenByRelaysTable::Table)
                    .columns([
                        EventSeenByRelaysTable::EventId,
                        EventSeenByRelaysTable::RelayURL,
                    ])
                    .values_panic([event_id.into(), relay_url.into()])
                    .build_rusqlite(SqliteQueryBuilder);

                match conn.execute(sql.as_str(), &*values.as_params()) {
                    Ok(updated_rows) => Ok(updated_rows),
                    Err(err) => Err(Error::internal(err.into())),
                }
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        if let Ok(event_id_inserted) = query_result {
            if event_id_inserted > 0 {
                return Ok(());
            }

            Err(Error::internal_with_message("Failed to insert new record"))
        } else {
            Err(Error::internal_with_message("Failed to insert new record"))
        }
    }

    async fn get_event_seen_on_relays(
        &self,
        event_id: EventId,
    ) -> Result<Option<HashSet<Url>>, Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };
        let event_id = event_id.to_string();

        let Ok(query_result) = connection
            .interact(|conn| -> Result<Option<HashSet<Url>>, Error> {
                let (sql, values) = Query::select()
                    .from(EventSeenByRelaysTable::Table)
                    .column(EventSeenByRelaysTable::RelayURL)
                    .and_where(sea_query::Expr::col(EventSeenByRelaysTable::EventId).eq(event_id))
                    .build_rusqlite(SqliteQueryBuilder);

                let mut stmt = conn.prepare(sql.as_str()).unwrap();
                let mut rows = stmt.query(&*values.as_params()).unwrap();

                let mut urls: HashSet<Url> = HashSet::new();
                while let Ok(Some(record)) = rows.next() {
                    if let Ok(url_string) = record.get::<_, String>("relay_url") {
                        if let Ok(relay_url) = Url::from_str(&url_string) {
                            urls.insert(relay_url);
                        }
                    }
                }

                if !urls.is_empty() {
                    return Ok(Some(urls));
                }

                Ok(None)
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        let Ok(found_relay_urls) = query_result else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        Ok(found_relay_urls)
    }

    async fn has_event_already_been_seen(&self, event_id: &EventId) -> Result<bool, Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let event_id = event_id.to_string();

        let Ok(query_result) = connection
            .interact(|conn| -> Result<bool, Error> {
                let (sql, values) = Query::select()
                    .expr_as(
                        sea_query::Expr::exists(
                            Query::select()
                                .column(EventSeenByRelaysTable::EventId)
                                .from(EventSeenByRelaysTable::Table)
                                .and_where(
                                    sea_query::Expr::col(EventSeenByRelaysTable::EventId)
                                        .eq(event_id),
                                )
                                .limit(1)
                                .take(),
                        ),
                        sea_query::Alias::new("event_exists"),
                    )
                    .build_rusqlite(SqliteQueryBuilder);

                let mut stmt = conn.prepare(sql.as_str()).unwrap();
                let mut rows = stmt.query(&*values.as_params()).unwrap();

                let exists = match rows.next() {
                    Ok(_) => 1,
                    Err(_) => 0,
                };

                Ok(exists == 1)
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        if let Ok(result) = query_result {
            return Ok(result);
        }

        Ok(false)
    }

    async fn admin_delete_events(&self, event: &Event) -> Result<RelayMessage, Error> {
        let event_ids: Vec<String> = event.event_ids().map(|e| e.to_string()).collect();
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };
        let Ok(query_result) = connection
            .interact(
                |conn: &mut rusqlite::Connection| -> Result<RelayMessage, Error> {
                    let (sql, values) = Query::delete()
                        .from_table(EventsTable::Table)
                        .and_where(sea_query::Expr::col(EventsTable::EventId).is_in(event_ids))
                        .build_rusqlite(SqliteQueryBuilder);

                    match conn.execute(sql.as_str(), &*values.as_params()) {
                        Ok(affected_rows) => {
                            let message = format!("{} events deleted", affected_rows);
                            Ok(RelayMessage::notice(message))
                        }
                        Err(err) => {
                            log::error!("[admin_delete_events] Failed to execute query: {}", err);

                            Ok(RelayMessage::notice("unable to delete events"))
                        }
                    }
                },
            )
            .await
        else {
            log::error!("[admin_delete_events] Failed to execute query");
            return Err(Error::internal_with_message(
                "Failed to execute query 'admin_delete_events'",
            ));
        };

        let Ok(relay_message) = query_result else {
            let message = RelayMessage::Notice {
                message: "unable to acquire pool connection".to_string(),
            };
            return Ok(message);
        };

        Ok(relay_message)
    }

    async fn get_event_by_id(&self, event_id: EventId) -> Result<Event, Error> {
        let eid = event_id.to_string();
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };
        let Ok(query_result) = connection
            .interact(|conn: &mut rusqlite::Connection| -> Result<Event, Error> {
                let (sql, value) = Query::select()
                    .from(EventsTable::Table)
                    .columns([
                        EventsTable::EventId,
                        EventsTable::Content,
                        EventsTable::Kind,
                        EventsTable::Pubkey,
                        EventsTable::CreatedAt,
                        EventsTable::Tags,
                        EventsTable::Sig,
                    ])
                    .and_where(sea_query::Expr::col(EventsTable::EventId).eq(eid))
                    .limit(1)
                    .build_rusqlite(SqliteQueryBuilder);

                let mut stmt = conn.prepare(sql.as_str()).unwrap();
                let row = stmt
                    .query_row::<EventRow, _, _>(&*value.as_params(), |row| Ok(EventRow::from(row)))
                    .unwrap();

                let event = row.clone().to_event();

                Ok(event)
            })
            .await
        else {
            return Err(Error::internal_with_message(
                "Failed to execute query 'get_event_by_id'",
            ));
        };

        let Ok(event) = query_result else {
            return Err(Error::internal_with_message(
                "Failed to get event from row in 'get_event_by_id'",
            ));
        };

        Ok(event)
    }

    async fn wipe(&self) -> Result<(), Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        match connection
            .interact(|conn| -> Result<(), Error> {
                let Ok(result) = conn.set_db_config(
                    rusqlite::config::DbConfig::SQLITE_DBCONFIG_RESET_DATABASE,
                    true,
                ) else {
                    return Err(Error::internal_with_message(
                        "Failed to set SQLITE_DBCONFIG_RESET_DATABASE to true",
                    ));
                };
                let Ok(result) = conn.execute("VACUUM;", []) else {
                    return Err(Error::internal_with_message("Failed to run VACUUM command"));
                };
                let Ok(result) = conn.set_db_config(
                    rusqlite::config::DbConfig::SQLITE_DBCONFIG_RESET_DATABASE,
                    false,
                ) else {
                    return Err(Error::internal_with_message(
                        "Failed to set SQLITE_DBCONFIG_RESET_DATABASE to false",
                    ));
                };

                Ok(())
            })
            .await
        {
            Ok(_) => {
                if NostrSqlite::run_migrations(&self.pool).await {
                    return Ok(());
                }

                Err(Error::internal_with_message("Failed to run migrations"))
            }
            Err(err) => Err(Error::internal_with_message(err.to_string())),
        }
    }

    fn get_filter_query(&self, filter: &nostr::Filter) -> sea_query::SelectStatement {
        let mut query = Query::select()
            .column((EventsTable::Table, EventsTable::EventId))
            .column((EventsTable::Table, EventsTable::Content))
            .columns([
                EventsTable::Kind,
                EventsTable::Pubkey,
                EventsTable::CreatedAt,
                EventsTable::Tags,
                EventsTable::Sig,
            ])
            .from(EventsTable::Table)
            .order_by(EventsTable::CreatedAt, sea_query::Order::Desc)
            .to_owned();

        if !filter.ids.is_empty() {
            let ids = filter.ids.iter().map(|id| id.to_string());
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::EventId)).is_in(ids),
                )
                .to_owned();
        }

        if !filter.kinds.is_empty() {
            let kinds: Vec<u32> = filter.kinds.iter().map(|kind| kind.as_u32()).collect();
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::Kind)).is_in(kinds),
                )
                .to_owned();
        }

        if !filter.authors.is_empty() {
            let authors = filter.authors.iter().map(|author| author.to_string());
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::Pubkey)).is_in(authors),
                )
                .to_owned();
        }

        if let Some(since) = filter.since {
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::CreatedAt))
                        .gte(since.as_u64()),
                )
                .to_owned();
        }

        if let Some(until) = filter.until {
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::CreatedAt))
                        .lte(until.as_u64()),
                )
                .to_owned();
        }

        if let Some(limit) = filter.limit {
            query = query.limit(limit as u64).to_owned();
        }

        filter.generic_tags.iter().for_each(|(tag, values)| {
            let values = values.iter().map(|val| val.to_string());
            query = query
                .left_join(
                    TagsTable::Table,
                    sea_query::Expr::col((TagsTable::Table, TagsTable::EventId))
                        .equals((EventsTable::Table, EventsTable::EventId)),
                )
                .and_where(
                    sea_query::Expr::col((TagsTable::Table, TagsTable::Tag)).eq(tag.to_string()),
                )
                .and_where(sea_query::Expr::col((TagsTable::Table, TagsTable::Value)).is_in(values))
                .to_owned();
        });

        if let Some(search) = &filter.search {
            query = query
                .inner_join(
                    EventsFTSTable::Table,
                    sea_query::Expr::col((EventsTable::Table, EventsTable::EventId))
                        .equals((EventsFTSTable::Table, EventsFTSTable::EventId)),
                )
                .and_where(
                    sea_query::Expr::col((EventsFTSTable::Table, EventsFTSTable::Content))
                        .matches(search),
                )
                .to_owned();
        }

        query
    }

    fn get_filters_query(
        &self,
        filters: Vec<Filter>,
        order: SqOrder,
    ) -> Option<sea_query::SelectStatement> {
        filters
            .iter()
            .map(|filter| {
                Query::select()
                    .column((EventsTable::Table, EventsTable::EventId))
                    .column((EventsTable::Table, EventsTable::Content))
                    .columns([
                        EventsTable::Kind,
                        EventsTable::Pubkey,
                        EventsTable::CreatedAt,
                        EventsTable::Tags,
                        EventsTable::Sig,
                    ])
                    .from_subquery(
                        self.get_filter_query(filter),
                        sea_query::Alias::new("events"),
                    )
                    .order_by(EventsTable::CreatedAt, order.to_owned())
                    .to_owned()
            })
            .reduce(|mut result, query| result.union(sea_query::UnionType::All, query).to_owned())
    }

    async fn query(&self, filters: Vec<Filter>, order: Order) -> Result<Vec<Event>, Error> {
        log::debug!("making query from filters...");

        // let event_ids: Vec<String> = event.event_ids().map(|e| e.to_string()).collect();
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let sq_order = match order {
            Order::Asc => SqOrder::Asc,
            Order::Desc => SqOrder::Desc,
        };
        let Some(sql_statement) = self.get_filters_query(filters, sq_order) else {
            return Err(Error::internal_with_message("Failed to build SQL Query"));
        };

        let Ok(query_result) = connection
            .interact(move |conn| -> Result<Vec<Event>, Error> {
                let (sql, values) = sql_statement.build_rusqlite(SqliteQueryBuilder);

                let mut stmt = conn.prepare(sql.as_str()).unwrap();
                let mut rows = stmt.query(&*values.as_params()).unwrap();

                let mut event_vec: Vec<Event> = vec![];
                while let Ok(Some(row)) = rows.next() {
                    let event = EventRow::from(row).to_event();
                    event_vec.push(event);
                }

                Ok(event_vec)
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        query_result
    }

    async fn count(&self, filters: Vec<Filter>) -> Result<usize, Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let Some(mut sql_statement) = self.get_filters_query(filters, SqOrder::Desc) else {
            return Err(Error::internal_with_message("Failed to build SQL Query"));
        };

        let Ok(query_result) = connection
            .interact(move |conn| {
                let (sql, values) = sql_statement
                    .clear_selects()
                    .expr_as(
                        sea_query::Func::count(sea_query::Expr::col((
                            EventsTable::Table,
                            EventsTable::EventId,
                        ))),
                        sea_query::Alias::new("count"),
                    )
                    .build_rusqlite(SqliteQueryBuilder);

                if let Ok(result) = conn.query_row(sql.as_str(), &*values.as_params(), |row| {
                    let count: usize = row.get(0).unwrap();
                    Ok(count)
                }) {
                    return Ok(result);
                }

                Err(Error::internal_with_message(
                    "Failed to get event counts by filters",
                ))
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        query_result
    }

    async fn event_ids_by_filters(
        &self,
        filters: Vec<Filter>,
        order: Order,
    ) -> Result<Vec<EventId>, Error> {
        let Ok(connection) = self.get_connection().await else {
            return Err(Error::internal_with_message("Unable to get DB connection"));
        };

        let sq_order = match order {
            Order::Asc => SqOrder::Asc,
            Order::Desc => SqOrder::Desc,
        };

        let Some(mut sql_statement) = self.get_filters_query(filters, sq_order) else {
            return Err(Error::internal_with_message("Failed to build SQL Query"));
        };

        let Ok(query_result) = connection
            .interact(move |conn| {
                let (sql, values) = sql_statement
                    .clear_selects()
                    .column(EventsTable::EventId)
                    .build_rusqlite(SqliteQueryBuilder);

                let mut stmt = conn.prepare(sql.as_str()).unwrap();
                let mut rows = stmt.query(&*values.as_params()).unwrap();

                let mut event_vec: Vec<EventId> = vec![];
                while let Ok(Some(row)) = rows.next() {
                    let event_id_string: String = row.get(0).unwrap();
                    let event_id = EventId::from_str(&event_id_string).unwrap();
                    event_vec.push(event_id);
                }

                Ok(event_vec)
            })
            .await
        else {
            return Err(Error::internal_with_message("Failed to execute query"));
        };

        query_result
    }

    async fn has_coordinate_been_deleted(
        &self,
        coordinate: &Coordinate,
        timestamp: Timestamp,
    ) -> Result<bool, Error> {
        Ok(true)
    }
}

impl From<nostr_database::DatabaseError> for Error {
    fn from(value: nostr_database::DatabaseError) -> Self {
        Error::internal_with_message(value.to_string())
    }
}

impl From<Error> for nostr_database::DatabaseError {
    fn from(val: Error) -> Self {
        nostr_database::DatabaseError::backend(val)
    }
}

#[async_trait]
impl NostrDatabase for NostrSqlite {
    type Err = Error;

    /// Name of the backend database used (ex. lmdb, sqlite, indexeddb, ...)
    fn backend(&self) -> Backend {
        Backend::SQLite
    }

    /// Database options
    fn opts(&self) -> DatabaseOptions {
        DatabaseOptions { events: true }
    }

    /// Save [`Event`] into store
    ///
    /// Return `true` if event was successfully saved into database.
    ///
    /// **This method assume that [`Event`] was already verified**
    async fn save_event(&self, event: &Event) -> Result<bool, Self::Err> {
        self.save_event(event).await
    }

    /// Check if [`Event`] has already been saved
    async fn has_event_already_been_saved(&self, event_id: &EventId) -> Result<bool, Self::Err> {
        self.has_event_already_been_saved(event_id).await
    }

    /// Check if [`EventId`] has already been seen
    async fn has_event_already_been_seen(&self, event_id: &EventId) -> Result<bool, Self::Err> {
        self.has_event_already_been_seen(event_id).await
    }

    /// Check if [`EventId`] has been deleted
    async fn has_event_id_been_deleted(&self, event_id: &EventId) -> Result<bool, Self::Err> {
        self.has_event_id_been_deleted(event_id).await
    }

    /// Check if event with [`Coordinate`] has been deleted before [`Timestamp`]
    async fn has_coordinate_been_deleted(
        &self,
        coordinate: &Coordinate,
        timestamp: Timestamp,
    ) -> Result<bool, Self::Err> {
        todo!()
    }

    /// Set [`EventId`] as seen by relay
    ///
    /// Useful for NIP65 (aka gossip)
    async fn event_id_seen(&self, event_id: EventId, relay_url: Url) -> Result<(), Self::Err> {
        self.set_event_id_seen(event_id, relay_url).await
    }

    /// Get list of relays that have seen the [`EventId`]
    async fn event_seen_on_relays(
        &self,
        event_id: EventId,
    ) -> Result<Option<HashSet<Url>>, Self::Err> {
        self.get_event_seen_on_relays(event_id).await
    }

    /// Get [`Event`] by [`EventId`]
    async fn event_by_id(&self, event_id: EventId) -> Result<Event, Self::Err> {
        self.get_event_by_id(event_id).await
    }

    /// Count number of [`Event`] found by filters
    ///
    /// Use `Filter::new()` or `Filter::default()` to count all events.
    async fn count(&self, filters: Vec<Filter>) -> Result<usize, Self::Err> {
        self.count(filters).await
    }

    /// Query store with filters
    async fn query(&self, filters: Vec<Filter>, order: Order) -> Result<Vec<Event>, Self::Err> {
        self.query(filters, order).await
    }

    /// Get event IDs by filters
    async fn event_ids_by_filters(
        &self,
        filters: Vec<Filter>,
        order: Order,
    ) -> Result<Vec<EventId>, Self::Err> {
        self.event_ids_by_filters(filters, order).await
    }

    /// Get `negentropy` items
    async fn negentropy_items(
        &self,
        filter: Filter,
    ) -> Result<Vec<(EventId, Timestamp)>, Self::Err> {
        todo!()
    }

    /// Wipe all data
    async fn wipe(&self) -> Result<(), Self::Err> {
        self.wipe().await
    }
}

impl Noose for NostrSqlite {
    async fn start(&mut self, pubsub: Arc<PubSub>) -> Result<(), Error> {
        let mut subscriber = pubsub.subscribe(channels::MSG_NOOSE).await;

        while let Ok(message) = subscriber.recv().await {
            log::info!("[Noose] received message: {:?}", message);
            let command = match message.content {
                Command::DbReqWriteEvent(client_id, event) => match self.write_event(event).await {
                    Ok(status) => Command::DbResOkWithStatus(client_id, status),
                    Err(e) => Command::ServiceError(e),
                },
                Command::DbReqFindEvent(client_id, subscriptioin) => {
                    match self.find_event(subscriptioin).await {
                        Ok(relay_messages) => {
                            Command::DbResRelayMessages(client_id, relay_messages)
                        }
                        Err(e) => Command::ServiceError(e),
                    }
                }
                Command::DbReqEventCounts(client_id, subscriptioin) => {
                    match self.counts(subscriptioin).await {
                        Ok(relay_message) => Command::DbResEventCounts(client_id, relay_message),
                        Err(e) => Command::ServiceError(e),
                    }
                }
                _ => Command::Noop,
            };
            if command != Command::Noop {
                let channel = message.source;
                let message = Message {
                    source: channels::MSG_NOOSE,
                    content: command,
                };

                log::info!(
                    "[Noose] publishing new message: {:?} to channel {}",
                    message,
                    channel
                );

                pubsub.publish(channel, message).await;
            }
        }

        Ok(())
    }

    async fn write_event(&self, event: Box<nostr::Event>) -> Result<nostr::RelayMessage, Error> {
        // TODO: Maybe do event validation and admin deletions here
        match self.save_event(&event).await {
            Ok(status) => {
                let relay_message = nostr::RelayMessage::ok(event.id, status, "");
                Ok(relay_message)
            }
            Err(err) => Err(Error::bad_request(err.to_string())),
        }
    }

    async fn find_event(
        &self,
        subscription: Subscription,
    ) -> Result<Vec<nostr::RelayMessage>, Error> {
        match self.query(subscription.filters, Order::Desc).await {
            Ok(events) => {
                let relay_messages = events
                    .into_iter()
                    .map(|event| nostr::RelayMessage::event(subscription.id.clone(), event))
                    .collect();
                Ok(relay_messages)
            }
            Err(err) => Err(Error::bad_request(err.to_string())),
        }
    }

    async fn counts(&self, subscription: Subscription) -> Result<nostr::RelayMessage, Error> {
        match self.count(subscription.filters).await {
            Ok(counts) => {
                let relay_message = nostr::RelayMessage::count(subscription.id, counts);
                Ok(relay_message)
            }
            Err(err) => Err(Error::internal_with_message(err.to_string())),
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::noose::sqlite::*;
    use nostr::EventBuilder;

    #[tokio::test]
    async fn create_db() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        assert_eq!(db.pool.status().max_size, 64);
    }

    #[tokio::test]
    async fn event_row_to_event() {
        let event_id =
            "076b27df50401f2af598cb727281e1e36401592570052c0fb37543f50aba5b69".to_string();
        let event_row = EventRow {
            id: event_id.clone(),
            pubkey: "da9cfc5b7644aae5b3e7df1d224ef4d4a206d66d6e3b5c99a8aeda3b642c4cef".to_string(),
            created_at: 1706020262,
            kind: 1,
            tags: "[]".to_string(),
            sig: "7f92a6ebdeae4a9bd7a6e19f57b49e2c499fece644c63c6ff5776bdbafec5c524cfad1eecedef11a8701df2a826453914d39a897e0f8276a6ffc40e266cae059".to_string(),
            content: "hoh".to_string(),
        };

        let event = event_row.to_event();

        assert_eq!(event.id.to_string(), event_id);
    }

    #[tokio::test]
    async fn get_event_by_id() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event = EventBuilder::new(nostr::Kind::TextNote, "Hello", vec![])
            .to_event(&keys)
            .unwrap();

        // Insert event
        let result = db.save_event(&event).await.unwrap();
        assert!(result);

        // Get event by id
        let eid = event.id;
        let result = db.get_event_by_id(eid).await.unwrap();

        assert_eq!(result.id, eid);
    }

    #[tokio::test]
    async fn event_seen_by_other_relays() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event = EventBuilder::new(nostr::Kind::TextNote, "Hello to Other relay", vec![])
            .to_event(&keys)
            .unwrap();

        // Insert event
        let result = db.save_event(&event).await.unwrap();
        assert!(result);

        // Set seen on other relays
        let url_0 = Url::from_str("wss://relay.zhitno.st").unwrap();
        db.set_event_id_seen(event.id, url_0).await.unwrap();

        let url_1 = Url::from_str("wss://relay.damus.io").unwrap();
        db.set_event_id_seen(event.id, url_1).await.unwrap();

        // Get status of seen event
        let result = db.get_event_seen_on_relays(event.id).await.unwrap();

        dbg!(result);

        // Get event by id
        let eid = event.id;
        let result = db.get_event_by_id(eid).await.unwrap();

        assert_eq!(result.id, eid);
    }

    #[tokio::test]
    async fn has_event_already_been_seen() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event = EventBuilder::new(nostr::Kind::TextNote, "Hello to Other relay", vec![])
            .to_event(&keys)
            .unwrap();

        // Set seen on other relays
        let url_0 = Url::from_str("wss://relay.zhitno.st").unwrap();
        db.set_event_id_seen(event.id, url_0).await.unwrap();

        let url_1 = Url::from_str("wss://relay.damus.io").unwrap();
        db.set_event_id_seen(event.id, url_1).await.unwrap();

        // Get status of seen event
        let result = db.has_event_already_been_seen(&event.id).await.unwrap();

        assert!(result);
    }

    #[tokio::test]
    async fn wipe_db() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        db.wipe().await.unwrap();
    }

    #[tokio::test]
    async fn query() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event = EventBuilder::new(nostr::Kind::TextNote, "Hello Filters", vec![])
            .to_event(&keys)
            .unwrap();

        // Insert event
        let result = db.save_event(&event).await.unwrap();
        assert!(result);

        // Get event vec from filters
        let filter = Filter::new().author(event.pubkey);
        let filters = vec![filter];
        let result = db.query(filters, Order::Desc).await.unwrap();

        assert!(!result.is_empty());
        assert_eq!(result[0].pubkey, event.pubkey);
    }

    #[tokio::test]
    async fn count() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event_0 = EventBuilder::new(nostr::Kind::TextNote, "Hello Count", vec![])
            .to_event(&keys)
            .unwrap();
        let event_1 = EventBuilder::new(nostr::Kind::TextNote, "Goodbye Count", vec![])
            .to_event(&keys)
            .unwrap();

        // Insert events
        let result = db.save_event(&event_0).await.unwrap();
        assert!(result);
        let result = db.save_event(&event_1).await.unwrap();
        assert!(result);

        // Get event vec from filters
        let filter = Filter::new().author(event_0.pubkey);
        let filters = vec![filter];
        let result = db.count(filters).await.unwrap();

        assert_eq!(result, 2);
    }

    #[tokio::test]
    async fn event_ids_by_filter() {
        let config = Arc::new(ServiceConfig::new());
        let db = NostrSqlite::new(config).await;

        let keys = nostr::Keys::generate();
        let event_0 = EventBuilder::new(nostr::Kind::TextNote, "Hello Count", vec![])
            .to_event(&keys)
            .unwrap();
        let event_1 = EventBuilder::new(nostr::Kind::TextNote, "Goodbye Count", vec![])
            .to_event(&keys)
            .unwrap();

        // Insert events
        let result = db.save_event(&event_0).await.unwrap();
        assert!(result);
        let result = db.save_event(&event_1).await.unwrap();
        assert!(result);

        // Get event vec from filters
        let filter = Filter::new().author(event_0.pubkey);
        let filters = vec![filter];
        let result = db.event_ids_by_filters(filters, Order::Desc).await.unwrap();

        assert_eq!(result.len(), 2);
    }
}
